@@ -20,6 +20,7 @@ import (
 type backupOpts struct {
 	name             string
 	typ              string
+	base             bool
 	compression      string
 	compressionLevel []int
 	ns               string
@@ -87,6 +88,7 @@ func runBackup(cn *pbm.PBM, b *backupOpts, outf outFormat) (fmt.Stringer, error)
 		Cmd: pbm.CmdBackup,
 		Backup: &pbm.BackupCmd{
 			Type:             pbm.BackupType(b.typ),
+			IncrBase:         b.base,
 			Name:             b.name,
 			Namespaces:       nss,
 			Compression:      compression,
@@ -280,14 +282,18 @@ func describeBackup(cn *pbm.PBM, b *descBcp) (fmt.Stringer, error) {
 	}
 
 	if bcp.Size == 0 {
-		stg, err := cn.GetStorage(cn.Logger().NewEvent("", "", "", primitive.Timestamp{}))
-		if err != nil {
-			return nil, errors.WithMessage(err, "get storage")
-		}
+		switch bcp.Status {
+		case pbm.StatusDone, pbm.StatusCancelled, pbm.StatusError:
+			stg, err := cn.GetStorage(cn.Logger().NewEvent("", "", "", primitive.Timestamp{}))
+			if err != nil {
+				return nil, errors.WithMessage(err, "get storage")
+			}
 
-		rv.Size, err = getLegacySnapshotSize(bcp.Replsets, bcp.Type, stg)
-		if err != nil {
-			return nil, errors.WithMessage(err, "get snapshot size")
+			rv.Size, err = getLegacySnapshotSize(bcp, stg)
+			if errors.Is(err, errMissedFile) && bcp.Status != pbm.StatusDone {
+				// canceled/failed backup can be incomplete. ignore
+				return nil, errors.WithMessage(err, "get snapshot size")
+			}
 		}
 	}
 
