@@ -109,8 +109,7 @@ func (n *Node) connect(direct bool) (*mongo.Client, error) {
 }
 
 func (n *Node) GetInfo() (*NodeInfo, error) {
-	i := &NodeInfo{}
-	err := n.cn.Database(DB).RunCommand(n.ctx, bson.D{{"isMaster", 1}}).Decode(i)
+	i, err := GetNodeInfo(n.ctx, n.cn)
 	if err != nil {
 		return nil, errors.Wrap(err, "get NodeInfo")
 	}
@@ -273,8 +272,7 @@ func (n *Node) WaitForWrite(ts primitive.Timestamp) (err error) {
 }
 
 func LastWrite(cn *mongo.Client, majority bool) (primitive.Timestamp, error) {
-	inf := &NodeInfo{}
-	err := cn.Database("admin").RunCommand(context.Background(), bson.D{{"isMaster", 1}}).Decode(inf)
+	inf, err := GetNodeInfo(context.TODO(), cn)
 	if err != nil {
 		return primitive.Timestamp{}, errors.Wrap(err, "get NodeInfo data")
 	}
@@ -366,14 +364,7 @@ func (n *Node) GetOpts(defaults *MongodOpts) (*MongodOpts, error) {
 }
 
 func (n *Node) GetRSconf() (*RSConfig, error) {
-	rsc := struct {
-		Conf RSConfig `bson:"config" json:"config"`
-	}{}
-	err := n.cn.Database("admin").RunCommand(n.ctx, bson.D{{"replSetGetConfig", 1}}).Decode(&rsc)
-	if err != nil {
-		return nil, errors.Wrap(err, "run mongo command")
-	}
-	return &rsc.Conf, nil
+	return GetReplSetConfig(n.ctx, n.cn)
 }
 
 func (n *Node) Shutdown() error {
@@ -383,3 +374,28 @@ func (n *Node) Shutdown() error {
 	}
 	return err
 }
+
+func GetNodeInfo(ctx context.Context, m *mongo.Client) (*NodeInfo, error) {
+	res := m.Database(DB).RunCommand(ctx, bson.D{{"isMaster", 1}})
+	if err := res.Err(); err != nil {
+		return nil, errors.WithMessage(err, "cmd: isMaster")
+	}
+
+	n := &NodeInfo{}
+	err := res.Decode(&n)
+	return n, errors.WithMessage(err, "decode")
+}
+
+func GetReplSetConfig(ctx context.Context, m *mongo.Client) (*RSConfig, error) {
+	res := m.Database("admin").RunCommand(ctx, bson.D{{"replSetGetConfig", 1}})
+	if err := res.Err(); err != nil {
+		return nil, errors.WithMessage(err, "run command")
+	}
+
+	val := struct{ Config *RSConfig }{}
+	if err := res.Decode(&val); err != nil {
+		return nil, errors.WithMessage(err, "decode")
+	}
+
+	return val.Config, nil
+}
